#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import gzip
import ssl
import time
import warnings
from typing import Any, Dict, Optional, Tuple, Union

import urllib3  # type: ignore[import]
from urllib3.exceptions import (  # type: ignore[import]
    ConnectTimeoutError,
    ReadTimeoutError,
)
from urllib3.util.retry import Retry  # type: ignore[import]

from .._compat import warn_stacklevel
from .._exceptions import ConnectionError, ConnectionTimeout, SecurityWarning, TlsError
from .._models import ApiResponseMeta, HttpHeaders, NodeConfig
from ..client_utils import DEFAULT, DefaultType, client_meta_version
from ._base import (
    BUILTIN_EXCEPTIONS,
    DEFAULT_CA_CERTS,
    RERAISE_EXCEPTIONS,
    BaseNode,
    ssl_context_from_node_config,
)

try:
    from ._urllib3_chain_certs import HTTPSConnectionPool
except (ImportError, AttributeError):
    HTTPSConnectionPool = urllib3.HTTPSConnectionPool  # type: ignore[misc]


class Urllib3HttpNode(BaseNode):
    """Default synchronous node class using the ``urllib3`` library via HTTP"""

    _CLIENT_META_HTTP_CLIENT = ("ur", client_meta_version(urllib3.__version__))

    def __init__(self, config: NodeConfig):
        super().__init__(config)

        pool_class = urllib3.HTTPConnectionPool
        kw: Dict[str, Any] = {}

        if config.scheme == "https":
            pool_class = HTTPSConnectionPool
            ssl_context = ssl_context_from_node_config(config)

            kw.update(
                {
                    "ssl_context": ssl_context,
                    "assert_hostname": config.ssl_assert_hostname,
                    "assert_fingerprint": config.ssl_assert_fingerprint,
                }
            )

            # Convert all sentinel values to their actual default
            # values if not using an SSLContext.
            ca_certs = DEFAULT_CA_CERTS if config.ca_certs is None else config.ca_certs
            if config.verify_certs:
                if not ca_certs:
                    raise ValueError(
                        "Root certificates are missing for certificate "
                        "validation. Either pass them in using the ca_certs parameter or "
                        "install certifi to use it automatically."
                    )

                kw.update(
                    {
                        "cert_reqs": "CERT_REQUIRED",
                        "ca_certs": ca_certs,
                        "cert_file": config.client_cert,
                        "key_file": config.client_key,
                    }
                )
            else:
                kw["cert_reqs"] = "CERT_NONE"

                if config.ssl_show_warn:
                    warnings.warn(
                        f"Connecting to {self.base_url!r} using TLS with verify_certs=False is insecure",
                        stacklevel=warn_stacklevel(),
                        category=SecurityWarning,
                    )
                else:
                    urllib3.disable_warnings()

        self.pool = pool_class(
            config.host,
            port=config.port,
            timeout=urllib3.Timeout(total=config.request_timeout),
            maxsize=config.connections_per_node,
            block=False,
            **kw,
        )

    def perform_request(
        self,
        method: str,
        target: str,
        body: Optional[bytes] = None,
        headers: Optional[HttpHeaders] = None,
        request_timeout: Union[DefaultType, Optional[float]] = DEFAULT,
    ) -> Tuple[ApiResponseMeta, bytes]:

        start = time.time()
        try:
            kw = {}
            if request_timeout is not DEFAULT:
                kw["timeout"] = request_timeout

            request_headers = self._headers.copy()
            if headers:
                request_headers.update(headers)

            if body:
                if self._http_compress:
                    body = gzip.compress(body)
                    request_headers["content-encoding"] = "gzip"
            else:
                body = None

            response = self.pool.urlopen(
                method,
                target,
                body=body,
                retries=Retry(False),
                headers=request_headers,
                **kw,
            )
            response_headers = HttpHeaders(response.headers)
            data = response.data
            duration = time.time() - start

        except RERAISE_EXCEPTIONS:
            raise
        except Exception as e:
            if isinstance(e, (ConnectTimeoutError, ReadTimeoutError)):
                raise ConnectionTimeout(
                    "Connection timed out during request", errors=(e,)
                ) from None
            elif isinstance(e, (ssl.SSLError, urllib3.exceptions.SSLError)):
                raise TlsError(str(e), errors=(e,)) from None
            elif isinstance(e, BUILTIN_EXCEPTIONS):
                raise
            raise ConnectionError(str(e), errors=(e,)) from None

        return (
            ApiResponseMeta(
                node=self.config,
                duration=duration,
                http_version="1.1",
                status=response.status,
                headers=response_headers,
            ),
            data,
        )

    def close(self) -> None:
        """
        Explicitly closes connection
        """
        self.pool.close()



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

from typing import Optional, Tuple, Union

from .._models import ApiResponseMeta, HttpHeaders
from ..client_utils import DEFAULT, DefaultType
from ._base import BaseNode


class BaseAsyncNode(BaseNode):
    """Base class for Async HTTP node implementations"""

    async def perform_request(  # type: ignore[override]
        self,
        method: str,
        target: str,
        body: Optional[bytes] = None,
        headers: Optional[HttpHeaders] = None,
        request_timeout: Union[DefaultType, Optional[float]] = DEFAULT,
    ) -> Tuple[ApiResponseMeta, bytes]:
        raise NotImplementedError()  # pragma: nocover

    async def close(self) -> None:  # type: ignore[override]
        raise NotImplementedError()  # pragma: nocover



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import asyncio
import logging
import os
import ssl
from typing import ClassVar, Optional, Tuple, Union

from .._models import ApiResponseMeta, HttpHeaders, NodeConfig
from .._version import __version__
from ..client_utils import DEFAULT, DefaultType

logger = logging.getLogger("elastic_transport.node")

DEFAULT_CA_CERTS: Optional[str] = None
DEFAULT_USER_AGENT = f"elastic-transport-python/{__version__}"
RERAISE_EXCEPTIONS = (RecursionError, asyncio.CancelledError)
BUILTIN_EXCEPTIONS = (
    ValueError,
    KeyError,
    NameError,
    AttributeError,
    LookupError,
    AssertionError,
    IndexError,
    MemoryError,
    RuntimeError,
    SystemError,
    TypeError,
)

try:
    import certifi

    DEFAULT_CA_CERTS = certifi.where()
except ImportError:  # pragma: nocover
    pass


class BaseNode:
    """
    Class responsible for maintaining a connection to a node. It
    holds persistent node pool to it and it's main interface
    (``perform_request``) is thread-safe.

    :arg config: :class:`~elastic_transport.NodeConfig` instance
    """

    _CLIENT_META_HTTP_CLIENT: ClassVar[Tuple[str, str]]

    def __init__(self, config: NodeConfig):
        self._config = config
        self._headers: HttpHeaders = self.config.headers.copy()  # type: ignore[attr-defined]
        self.headers.setdefault("connection", "keep-alive")
        self.headers.setdefault("user-agent", DEFAULT_USER_AGENT)
        self._http_compress = bool(config.http_compress or False)
        if config.http_compress:
            self.headers["accept-encoding"] = "gzip"

        self._scheme = config.scheme
        self._host = config.host
        self._port = config.port
        self._path_prefix = (
            ("/" + config.path_prefix.strip("/")) if config.path_prefix else ""
        )

    @property
    def config(self) -> NodeConfig:
        return self._config

    @property
    def headers(self) -> HttpHeaders:
        return self._headers

    @property
    def scheme(self) -> str:
        return self._scheme

    @property
    def host(self) -> str:
        return self._host

    @property
    def port(self) -> int:
        return self._port

    @property
    def path_prefix(self) -> str:
        return self._path_prefix

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}({self.base_url})>"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseNode):
            return NotImplemented
        return self.__hash__() == other.__hash__()

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, BaseNode):
            return NotImplemented
        return not self == other

    def __hash__(self) -> int:
        return hash((str(type(self).__name__), self.config))

    @property
    def base_url(self) -> str:
        return "".join(
            [
                self.scheme,
                "://",
                # IPv6 must be wrapped by [...]
                "[%s]" % self.host if ":" in self.host else self.host,
                ":%s" % self.port if self.port is not None else "",
                self.path_prefix,
            ]
        )

    def perform_request(
        self,
        method: str,
        target: str,
        body: Optional[bytes] = None,
        headers: Optional[HttpHeaders] = None,
        request_timeout: Union[DefaultType, Optional[float]] = DEFAULT,
    ) -> Tuple[ApiResponseMeta, bytes]:  # pragma: nocover
        raise NotImplementedError()

    def close(self) -> None:  # pragma: nocover
        pass

    def log_request_success(self, method, url, body, status, response, duration):  # type: ignore
        """Log a successful API call"""
        # body has already been serialized to utf-8, deserialize it for logging
        # TODO: find a better way to avoid (de)encoding the body back and forth
        if body:
            try:
                body = body.decode("utf-8", "ignore")
            except AttributeError:
                pass

        logger.info("%s %s [status:%s request:%.3fs]", method, url, status, duration)
        logger.debug("> %s", body)
        logger.debug("< %s", response)

    def log_request_fail(  # type: ignore
        self,
        method,
        url,
        body,
        duration,
        status=None,
        response=None,
        exception=None,
    ):
        """Log an unsuccessful API call"""
        # do not log 404s on HEAD requests
        if method == "HEAD" and status == 404:
            return
        logger.warning(
            "%s %s [status:%s request:%.3fs]",
            method,
            url,
            status or "N/A",
            duration,
            exc_info=exception is not None,
        )

        # body has already been serialized to utf-8, deserialize it for logging
        # TODO: find a better way to avoid (de)encoding the body back and forth
        if body:
            try:
                body = body.decode("utf-8", "ignore")
            except AttributeError:
                pass

        logger.debug("> %s", body)

        if response is not None:
            logger.debug("< %s", response)


_HAS_TLS_VERSION = hasattr(ssl, "TLSVersion")
_SSL_PROTOCOL_VERSION_ATTRS = ("TLSv1", "TLSv1_1", "TLSv1_2")
_SSL_PROTOCOL_VERSION_DEFAULT = getattr(ssl, "OP_NO_SSLv2", 0) | getattr(
    ssl, "OP_NO_SSLv3", 0
)
_SSL_PROTOCOL_VERSION_TO_OPTIONS = {}
_SSL_PROTOCOL_VERSION_TO_TLS_VERSION = {}
for i, _protocol_attr in enumerate(_SSL_PROTOCOL_VERSION_ATTRS):
    try:
        _protocol_value = getattr(ssl, f"PROTOCOL_{_protocol_attr}")
    except AttributeError:
        continue

    if _HAS_TLS_VERSION:
        _tls_version_value = getattr(ssl.TLSVersion, _protocol_attr)
        _SSL_PROTOCOL_VERSION_TO_TLS_VERSION[_protocol_value] = _tls_version_value
        _SSL_PROTOCOL_VERSION_TO_TLS_VERSION[_tls_version_value] = _tls_version_value

    # Because we're setting a minimum version we binary OR all the options together.
    _SSL_PROTOCOL_VERSION_TO_OPTIONS[
        _protocol_value
    ] = _SSL_PROTOCOL_VERSION_DEFAULT | sum(
        getattr(ssl, f"OP_NO_{_attr}", 0) for _attr in _SSL_PROTOCOL_VERSION_ATTRS[:i]
    )

# TLSv1.3 is unique, doesn't have a PROTOCOL_TLSvX counterpart. So we have to set it manually.
if _HAS_TLS_VERSION:
    try:
        _SSL_PROTOCOL_VERSION_TO_TLS_VERSION[
            ssl.TLSVersion.TLSv1_3
        ] = ssl.TLSVersion.TLSv1_3
    except AttributeError:  # pragma: nocover
        pass


def ssl_context_from_node_config(node_config: NodeConfig) -> ssl.SSLContext:
    if node_config.ssl_context:
        ctx = node_config.ssl_context
    else:
        ctx = ssl.create_default_context()

        # Enable/disable certificate verification in these orders
        # to avoid 'ValueErrors' from SSLContext. We only do this
        # step if the user doesn't pass a preconfigured SSLContext.
        if node_config.verify_certs:
            ctx.verify_mode = ssl.CERT_REQUIRED
            ctx.check_hostname = True
        else:
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE

    # Enable logging of TLS session keys for use with Wireshark.
    if hasattr(ctx, "keylog_filename"):
        sslkeylogfile = os.environ.get("SSLKEYLOGFILE", "")
        if sslkeylogfile:
            ctx.keylog_filename = sslkeylogfile  # type: ignore[attr-defined]

    # Apply the 'ssl_version' if given, otherwise default to TLSv1.2+
    ssl_version = node_config.ssl_version
    if ssl_version is None:
        if _HAS_TLS_VERSION:
            ssl_version = ssl.TLSVersion.TLSv1_2
        else:
            ssl_version = ssl.PROTOCOL_TLSv1_2

    try:
        if _HAS_TLS_VERSION:
            ctx.minimum_version = _SSL_PROTOCOL_VERSION_TO_TLS_VERSION[ssl_version]
        else:
            ctx.options |= _SSL_PROTOCOL_VERSION_TO_OPTIONS[ssl_version]
    except KeyError:
        raise ValueError(
            f"Unsupported value for 'ssl_version': {ssl_version!r}. Must be "
            "either 'ssl.PROTOCOL_TLSvX' or 'ssl.TLSVersion.TLSvX'"
        ) from None

    return ctx



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

from typing import (
    Any,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    overload,
)

from ._models import ApiResponseMeta

_RawType = TypeVar("_RawType")
_BodyType = TypeVar("_BodyType")
_ObjectBodyType = TypeVar("_ObjectBodyType")
_ListItemRawType = TypeVar("_ListItemRawType")
_ListItemBodyType = TypeVar("_ListItemBodyType")


class ApiResponse(Generic[_RawType, _BodyType]):
    """Base class for all API response classes"""

    __slots__ = ("_raw", "_meta", "_body_cls")

    def __init__(
        self,
        raw: _RawType,
        meta: ApiResponseMeta,
        body_cls: Optional[Type[_BodyType]] = None,
    ):
        self._raw = raw
        self._meta = meta
        self._body_cls = body_cls

    def __repr__(self) -> str:
        body_repr = self.body
        if body_repr is None:
            body_repr = self.raw
        return f"{type(self).__name__}({body_repr!r})"

    def __contains__(self, item: Any) -> bool:
        return item in self._raw  # type: ignore[operator]

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ApiResponse):
            other = other._raw
        return self._raw == other

    def __ne__(self, other: object) -> bool:
        if isinstance(other, ApiResponse):
            other = other.raw
        return self._raw != other

    def __getitem__(self, item: Any) -> Any:
        return self._raw[item]  # type: ignore[index]

    def __getattr__(self, attr: str) -> Any:
        return getattr(self._raw, attr)

    def __len__(self) -> int:
        return len(self._raw)  # type: ignore[arg-type]

    def __iter__(self) -> Iterable[Any]:
        return iter(self._raw)  # type: ignore[no-any-return,call-overload]

    def __str__(self) -> str:
        return str(self._raw)

    def __bool__(self) -> bool:
        return bool(self._raw)

    @property
    def meta(self) -> ApiResponseMeta:
        """Response metadata"""
        return self._meta

    @property
    def raw(self) -> _RawType:
        """Raw deserialized response"""
        return self._raw

    @property
    def body(self) -> _BodyType:
        """User-friendly view into the raw response with type hints if applicable"""
        raise NotImplementedError()


class TextApiResponse(ApiResponse[str, str]):
    """API responses which are text such as 'text/plain' or 'text/csv'"""

    def __init__(self, raw: str, meta: ApiResponseMeta):
        super().__init__(raw=raw, meta=meta)

    def __iter__(self) -> Iterable[str]:
        return iter(self._raw)

    def __getitem__(self, item: Union[int, slice]) -> str:
        return self._raw[item]

    @property
    def raw(self) -> str:
        return self._raw

    @property
    def body(self) -> str:
        return self.raw


class BinaryApiResponse(ApiResponse[bytes, bytes]):
    """API responses which are a binary response such as Mapbox vector tiles"""

    def __init__(self, raw: bytes, meta: ApiResponseMeta):
        super().__init__(raw=raw, meta=meta)

    def __iter__(self) -> Iterable[int]:
        return iter(self.raw)

    @overload
    def __getitem__(self, item: slice) -> bytes:
        ...

    @overload
    def __getitem__(self, item: int) -> int:
        ...

    def __getitem__(self, item: Union[int, slice]) -> Union[int, bytes]:
        return self.raw[item]

    @property
    def raw(self) -> bytes:
        return self._raw

    @property
    def body(self) -> bytes:
        return self.raw


class HeadApiResponse(ApiResponse[bool, bool]):
    """API responses which are for an 'exists' / HEAD API request"""

    def __init__(self, meta: ApiResponseMeta):
        super().__init__(raw=200 <= meta.status < 300, meta=meta)

    def __bool__(self) -> bool:
        return 200 <= self.meta.status < 300

    @property
    def raw(self) -> bool:
        return bool(self)

    @property
    def body(self) -> bool:
        return bool(self)


class ObjectApiResponse(
    Generic[_ObjectBodyType], ApiResponse[Dict[str, Any], _ObjectBodyType]
):
    """API responses which are for a JSON object"""

    def __init__(self, raw: Dict[str, Any], meta: ApiResponseMeta):
        super().__init__(raw=raw, meta=meta)

    def __getitem__(self, item: str) -> Any:
        return self.raw[item]

    def __iter__(self) -> Iterator[str]:
        return iter(self.raw)

    @property
    def raw(self) -> Dict[str, Any]:
        return self._raw

    @property
    def body(self) -> _ObjectBodyType:
        if self._body_cls is None:
            return None  # type: ignore[return-value]
        return self._body_cls(self.raw)  # type: ignore[call-arg]


class ListApiResponse(
    Generic[_ListItemRawType, _ListItemBodyType],
    ApiResponse[List[_ListItemRawType], List[_ListItemBodyType]],
):
    """API responses which are a list of items. Can be NDJSON or a JSON list"""

    def __init__(self, raw: List[Any], meta: ApiResponseMeta):
        super().__init__(raw=raw, meta=meta)

    @overload
    def __getitem__(self, item: slice) -> List[_ListItemRawType]:
        ...

    @overload
    def __getitem__(self, item: int) -> _ListItemRawType:
        ...

    def __getitem__(
        self, item: Union[int, slice]
    ) -> Union[_ListItemRawType, List[_ListItemRawType]]:
        return self.raw[item]

    def __iter__(self) -> Iterable[_ListItemRawType]:
        return iter(self.raw)

    @property
    def raw(self) -> List[_ListItemRawType]:
        return self._raw

    @property
    def body(self) -> List[_ListItemBodyType]:
        if self._body_cls is None:
            return None  # type: ignore[return-value]
        return [self._body_cls(item) for item in self._raw]  # type: ignore[call-overload]
